
#include "eel.h"
#include "util.h"

#define ABB_BUF "-Abbrevs"
#define ABB_STR_LEN 81     /* Abbrev len w/ room for \0 at end */

#define EXPANDERS " \t\n\r!@#$%^&*()+-={}|[]\\:\";'<>?,./"

/* abbreviation mode - by Stephen Robbins  5/31/91
   
   This mode allows you to create abbreviations; type one thing, and another
   is inserted in the buffer.  Certain activation characters (space, \n, etc.)
   trigger expansion of the abbreviation before the point.  If the expansion
   contains a control-A, it is deleted, and the point is left there after
   expansion.  In that case, the expansion-triggering character isn't
   inserted afterwards.

   Note:  abbrev-mode is global; not buffer specific.  The reason is that
   keytables are shared between modes (e.g. c_tab is used by all c-mode
   buffers), and saving/restoring bindings for word abbrev mode gets very,
   very tricky if you want it to be per-buffer rather than per-mode.
   
   Commands:
   
   C-X +        Define an abbreviation for the word before the point.  Given a
                numeric argument, define the abbreviation to include the
                previous N words.
   C-X -        Define the word before the point as an abbreviation for 
                itself.  Useful for auto-capitalization with case sensitive
                languages.

   Extended commands:
   
   M-X abbrev-mode       Enter the mode where abbrevs are expanded and
                         defining them is available.
   M-X add-abbrev        Prompt for and add an abbreviation.
                         With numeric arg, adds the region as the expansion.
   M-X remove-abbrev     Prompt for and remove an abbreviation.
   M-X clear-all-abbrevs Clear all abbreviations.
   M-X load-abbrev-file  Load an abbreviation file
   M-X save-abbrevs      Save the current abbreviation file

   Changed commands:
   
   M-X exit              Exit offers to save unsaved word abbreviations.

   Implementation:
   
   Abbreviations are kept in the buffer ABB_BUF.  Since ; is an expander
   for a abb (and can't appear in the abbreviation), each abbreviation is:
        abbrev;expansion\n
   Newlines within the expansion are changed to ^Qs, and changed back when the
   expansion is inserted.  This is so each abbreviation is one logical "line"
   to epsilon, for sorting and nl_forward() purposes.

   Bugs:
   
   The whole key binding mess and the inability to have abbrev mode
   selectively on in buffers is a real hassle.  If Epsilon provided an
   arbitrary stack of key tables which we could use to temporarily mask
   certain keys on a buffer-by-buffer basis, that would be great.  Even so,
   entering and leaving modes could screw up the keytables.  Epsilon needs a
   minor-mode capability, which is beyond my patience at the moment.

   Enhancements:

   Case-hacking word abbrev, so:
        abc    ->   a big company
        Abc    ->   A Big Company
        ABC    ->   A BIG COMPANY
        
 */

/* There's a little bit of Lugaru code in here--at least the mode line
 * modification routine.
*************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*	"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.	*
*									*
*  Copyright (C) 1985, 1992 Lugaru Software Ltd.  All rights reserved.	*
*									*
* Limited permission is hereby granted to reproduce and modify this	*
* copyrighted material provided that the resulting code is used only in	*
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.				*
************************************************************************/

char abbrev_file[FNAMELEN] = "Abbrevs.abb";

buffer char abbrev_expanders[] = EXPANDERS;

/* these are not intended to be directly set */
int in_abbrev_mode = 0;                 /* Is Abbrev mode on? */
short saved_abb_keys[NUMKEYS];          /* Saved key functions */
short saved_cxplus;                     /* Saved C-X + function */
short saved_cxminus;                    /* Saved C-X - function */

init_abbs()
{
     char abb_file[FNAMELEN];

     if (!exist(ABB_BUF)) {
          strcpy(abb_file, abbrev_file);
          absolute(abb_file);
          load_abbs(abb_file);
          }
}

load_abbs(abb_file)
char *abb_file;
{
     char *oldbuf = bufname;
     int i;

     zap(ABB_BUF);
     bufname = ABB_BUF;
     i = read_file(abb_file, 1);
     point = 0;
     if (size() == 0 || curchar() != '\n') insert('\n');
     bufname = oldbuf;
     if (i && i != 2) {
          delete_buffer(ABB_BUF);
          quick_abort();
          }
}

do_save_abbs()		/* save abbreviations to their file if unsaved */
{
     char *oldbuf = bufname;

     if (exist(ABB_BUF)) {
          bufname = ABB_BUF;
          if (modified) {		/* saved tags must always be sorted */
               noteput ("Sorting Abbrevs...");
               sort_another(ABB_BUF,0,0);
               do_save_file(1, 1, 1);
               }
          bufname = oldbuf;
          }
}

add_abb(abbrev, expansion)
char *abbrev;
char *expansion;
{
     char *oldbuf = bufname;
     char src_tmp[ABB_STR_LEN+2];
     char msg[80];

     if (!*abbrev) error("Can't add a null abbreviation.");
     
     init_abbs();
     bufname = ABB_BUF;
     
     sprintf(src_tmp, "\n%s;", abbrev); /* Fine \n<xxx>;<old expand> */
     point = 0;
     change_chars(expansion, '\n', '\021'); /* Store ^Q for embedded \n's */
     if (search(1, src_tmp)) {
          int start = point;
          to_end_line();
          delete (start, point);
          stuff (expansion);
          sprintf(msg, "Replaced \"%s\" => \"", abbrev);
          }
     else {
          point = size();
          bprintf("%s;%s\n", abbrev, expansion);
          sprintf(msg, "Added \"%s\" => \"", abbrev);
          }
     strcat_printable (msg, expansion, 40);
     strcat (msg, "\"");
     noteput ("%s", msg);
     bufname = oldbuf;
}

make_abbrev_prompt (dest, first, second, third)
char *dest, *first, *second, *third;
{
     *dest = '\0';
     strcat_printable (dest, first,  20);
     strcat_printable (dest, second, 40);
     strcat_printable (dest, third,  20);
}
     
/*****
 ***
 *** Completion handling of abbreviations
 ***
 *****/

char abb_mat[ABB_STR_LEN]; /* value returned by abb_match during completion */

char *abb_match(partial, start) 
char *partial;
{
     char *oldbuf = bufname, *res = 0, tmp[ABB_STR_LEN+1];
     int found_len;

     bufname = ABB_BUF;

     if (start & STARTMATCH) {       /* Get the first abb match */
          if (modified) sort_another(ABB_BUF,0,0);
          point = 0;
          sprintf(tmp, "\n%s", partial);
          search (1, tmp);
          to_begin_line();
          }

     /* POINT is at start of next match line */
     if ((!*partial || parse_string(1, partial, NULL))
          && (found_len = parse_string(1, ".+;", abb_mat))) {
          abb_mat[found_len - 1] = 0;
          nl_forward();
          res = abb_mat;
          }
     bufname = oldbuf;
     return res;
}

get_abb(result, prompt)
char *result, *prompt;
{
     comp_read(result, prompt, abb_match, COMP_FOLD, "");
}

/*****
 ***
 *** User visible routines
 ***
 *****/

/* Point is assumed to be immediately following abbreviation, though
   separation by whitespace is OK. 
   
   Sets 'key' to \0 if the abbreviation leaves cursor not-at-end.
 */
command expand_abbrev_at_point()
{
     spot orig_point = alloc_spot();
     spot orig_mark  = alloc_spot();
     int orig_iter = iter;
     int orig_has_arg = has_arg;
     
     int abb_end = point, 
         expan_start;
     int len;                           /* Length of expansion */

     char *orig_buf = bufname, *expansion;
     char abb[ABB_STR_LEN];             /* abb= abbreviation */
     char tmp[ABB_STR_LEN+2];           /* tmp= search string */
     char *final_point;                 /* where in expansion is ^A */
     
     if (!exist(ABB_BUF)) return;       /* No abbrevs loaded- return */
     
     *orig_point = point;
     *orig_mark  = mark;
     
     has_arg = 0;
     iter = 1;
     
     backward_word();

     /* If word before point is too long, don't bother expanding. */
     if ((abb_end - point) >= ABB_STR_LEN) {
          point = *orig_point;
          goto done;
          }

     grab(point, abb_end, abb);
     bufname = ABB_BUF;
     point = 0;
     sprintf(tmp, "\n%s;", abb);
     if (search(1, tmp)) {              /* Found this abbreviation */
          expan_start = point;
          to_end_line();
          len = point - expan_start;
          expansion = malloc(len + 1);
          grab (expan_start, point, expansion);
          bufname = orig_buf;
          delete(point, abb_end);
          change_chars(expansion, '\021', '\n');
          stuff(expansion);
          if ((final_point = rindex(expansion, '\001')) != NULL) {
               point -= (len - (final_point - expansion));
               delete(point, point+1);
               key = '\0';              /* Don't insert char after expand */
               }
          free(expansion);
          }
     else {                             /* No such abbreviation */
          bufname = orig_buf;
          point = *orig_point;
          }
     
done:
     mark  = *orig_mark;
     iter  = orig_iter;
     has_arg = orig_has_arg;
     
     free_spot(orig_point);
     free_spot(orig_mark);
}

command save_abbrevs()
{
     do_save_abbs();
}

command load_abbrev_file()
{
     char abb_file[FNAMELEN];
     char *def_file;
     char *old = bufname;

     do_save_abbs();
     bufname = ABB_BUF;
     def_file = exist(ABB_BUF) ? filename : abbrev_file;
     bufname = old;
     get_file(abb_file, "Abbrev file", def_file);
     load_abbs(abb_file);
     say("Abbrevs loaded from %s", abb_file);
}

command clear_all_abbrevs()	/* erase all abbrevs */
{
     char *orig_buf = bufname;
     
     init_abbs();
     zap(ABB_BUF);
     bufname = ABB_BUF;
     insert('\n');
     bufname = orig_buf;
     do_save_abbs();
}

command add_abbrev()
{
     if (has_arg) add_abbrev_region();
     else add_abbrev_prompting();
     iter = 0;
}

add_abbrev_region()
{       
     char abbrev[ABB_STR_LEN], prompt[80], *tmp;
     int region_len = (mark > point) ? (mark-point+1) : (point-mark+1);
     
     /* This logic frees TMP before calling get_string because the user can
        type C-G and abort, without properly freeing TMP's storage. */
     if (mark != point) {
          tmp = malloc (region_len);
          grab (point, mark, tmp);
          make_abbrev_prompt (prompt, "Abbrev for \"", tmp, "\": ");
          free(tmp);
          get_string (abbrev, prompt);
          tmp = malloc (region_len);
          grab (point, mark, tmp);
          add_abb(abbrev, tmp);
          free(tmp);
          }
     else error("No region to define an abbreviation for.");
}

add_abbrev_prompting()
{
     char abbrev[ABB_STR_LEN], expansion[132];
     
     get_string(abbrev, "Abbrev: ");
     get_string(expansion, "Expansion: ");
     add_abb(abbrev, expansion);
}

/* add_abbrev_at_point()
   
   With numeric arg, make that many words preceding point be the expansion.
   Prompt for the abbreviation and install it. 
 */
command add_abbrev_at_point()           /* Position point at expansion */
{
     add_point_abbrev(FALSE);
}

/* add_self_abbrev_at_point()
   
   Add the single word immediately before (after if -1 arg) the point be its
   own expansion.  Useful for making case-insensitive abbreviations which
   expand to case-sensitive expansions.
 */
command add_self_abbrev_at_point()
{
     if (iter < 0) 
          iter = -1; 
     else
          iter = 1;
     add_point_abbrev(TRUE);
}
     
add_point_abbrev(is_self_abbrev)
int is_self_abbrev;
{
     char abbrev[ABB_STR_LEN], *expansion;
     char prompt[80];
     int l;
     int orig_point = point, expan_end, expan_start;
     
     if (iter > 0) {                    /* Storing word before point */
          re_search(-1, "[ \t\n]*");
          expan_end = point;
          backward_word();
          expan_start = point;
          }
     else {                             /* Storing words after point */
          re_search(1, "[ \t\n]*");
          expan_start = point;
          backward_word();
          expan_end   = point;
          }
          
     point = orig_point;
     l = expan_end - expan_start;
     expansion = malloc(l + 1);
     grab (expan_start, expan_end, expansion);

     if (is_self_abbrev) {
          add_abb(expansion, expansion);
          }
     else {
          make_abbrev_prompt (prompt, "Abbrev for \"", expansion, "\": ");
          get_string(abbrev, prompt);
          add_abb(abbrev, expansion);
          }

     free(expansion);
     iter = 0;
}

command remove_abbrev()
{
     char abbrev[ABB_STR_LEN], tmp[ABB_STR_LEN+2];

     if (!exist(ABB_BUF)) {
          noteput ("No abbrevs defined.");
          return;
          }
     
     get_abb(abbrev, "Remove abbrev: ");
     sprintf(tmp, "\n%s;", abbrev);
     bufname = ABB_BUF;
     point = 0;
     if (search(1, tmp)) {
          int start;
          to_begin_line();
          start = point;
          nl_forward();
          delete (start, point);
          noteput ("\"%s\" removed.", abbrev);
          }
     else noteput ("No such abbrev: \"%s\"", abbrev);
}


/*****
 ***
 *** Install abbrev expand on various characters
 ***
 *****/

/* char_expand_abbrev_at_point() - expand appreviation right before the point,
   then do this character's normal action.
 */

char_expand_abbrev_at_point()
{
     int orig_key = key;
     
     expand_abbrev_at_point();
     if (key != '\0') {
          root_keys[key] = saved_abb_keys[key];
          do_again();
          root_keys[orig_key] = (short) char_expand_abbrev_at_point;
          }
}

command abbrev_mode()
{
     char *c;

     if (!in_abbrev_mode) {
          for (c = abbrev_expanders; *c != '\0'; c++) {
               saved_abb_keys[*c] = root_keys[*c];
               root_keys[*c] = (short) char_expand_abbrev_at_point;
               }
          saved_cxplus = cx_tab['+'];
          saved_cxminus= cx_tab['-'];
          cx_tab['+'] = (short) add_abbrev_at_point;
          cx_tab['-'] = (short) add_self_abbrev_at_point;
          init_abbs();
          }
     else {
          for (c = abbrev_expanders; *c != '\0'; c++)
               root_keys[*c] = saved_abb_keys[*c];
          cx_tab['+'] = saved_cxplus;
          cx_tab['-'] = saved_cxminus;
          }
     in_abbrev_mode = !in_abbrev_mode;
     make_mode();
}

/* EPSILON 5.01:
new_make_mode()
{
     strcpy(mode, major_mode);
     if (fill_mode)
          strcat(mode, " Fill");
     if (over_mode)
          strcat(mode, " Over");
     if (in_rectangle_mode)
          strcat(mode, " Rect");
     if (strip_returns != strip_returns.default)
          strcat(mode, strip_returns ? " Trans" : " NoTrans");
     if (len_def_mac)
          strcat(mode, " Def");
     if (narrow_start > 0 || narrow_end > 0)
          strcat(mode, " Narrow");
     if (in_abbrev_mode)
          strcat(mode, " Abbrev");
}
/* END EPSILON 5.0 */

/* EPSILON 6.0 
new_set_mode(mode)	/* add major and minor modes to mode line 
char *mode;
{
	strcpy(mode, major_mode);
	if (fill_mode)
		strcat(mode, " Fill");
	if (over_mode)
		strcat(mode, " Over");
	if (in_rectangle_mode)
		strcat(mode, " Rect");
	if (strip_returns != strip_returns.default)
		strcat(mode, strip_returns ? " Trans" : " NoTrans");
	if (_len_def_mac)
		strcat(mode, " Def");
	if (narrow_start > 0 || narrow_end > 0)
		strcat(mode, " Narrow");
        if (in_abbrev_mode)
	       strcat(mode, " Abbrev");
	if (_buf_readonly)
		strcat(mode, " RO");
}
/* END EPSILON 6.0 */

/* EPSILON 7.0 */
new_set_mode(mode)	/* add major and minor modes to mode line */
char *mode;
{
	strcpy(mode, major_mode);
	if (fill_mode)
		strcat(mode, " Fill");
	if (over_mode)
		strcat(mode, " Over");
	if (strip_returns != strip_returns.default)
		strcat(mode, strip_returns ? " Trans" : " NoTrans");
	if (_len_def_mac)
		strcat(mode, " Def");
	if (ignore_kbd_macro || recording_suspended)
		strcat(mode, " Susp");
	if (narrow_start > 0 || narrow_end > 0)
		strcat(mode, " Narrow");
        if (in_abbrev_mode)
	       strcat(mode, " Abbrev");
	if (_buf_readonly)
		strcat(mode, " RO");
}
/* EPSILON 7.0 */

new_exit()
{
     char *orig_buf = bufname;
     char response[10];
     
     if (exist(ABB_BUF)) {
          bufname = ABB_BUF;
          if (modified) {
               bufname = orig_buf;      /* In case user hits <Abort> */
               get_string(response, "Word abbrevs not saved; save? [y]");
               if (tolower(*response) != 'n') do_save_abbs();
               else { 
                    bufname = ABB_BUF;
                    modified = FALSE;
                    }
               }
          }
     bufname = orig_buf;
     orig_abbrev_exit();
}


when_loading () /* Install new MAKE_MODE() unless we've already done so */
{

    if (!find_index("orig_abb_set_mode"))
	replace_name("set_mode", "orig_abb_set_mode");
    else
	drop_name("set_mode");
    replace_name("new_set_mode","set_mode");

    if (!find_index("orig_abbrev_exit"))
         replace_name("exit", "orig_abbrev_exit");
    else
         drop_name("exit");
    replace_name("new_exit", "exit");

}

